//
//  CB_Filter.java
//  CBSim
//
//  Created by Loren Petrich on 5/19/08.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

import java.awt.*;
import java.awt.image.*;

public abstract class CB_Filter extends RGBImageFilter
{
	// Methods to implement in its subclass:
	protected abstract void SetTransformType(int cbtype_);
	protected abstract void DoTransform(double[] ClrIn, double[] ClrOut);
	
	// Input parameters:
	
	// No internal-state changes, so get/set is overkill
	public double fraction = 1;
	public int off_edge = 0;
	
	// Has get/set, since changing it changes some internal state
	private int cbtype;
	private double gamma;
	
	// Forward and backward intensity-conversion lookup tables;
	// this is done to avoid calculating 6 Math.pow()'s per pixel.
	// Forward: int to double
	// Inverse: int (expanded) to int
	//
	// Inverse values are quantized to InvGamTblSize;
	// its size is chosen to balance precision and amount of precalculation;
	// the forward table requires much less precalculation.
	// Timed with System.nanoTime();
	// 65536 values takes about 0.03 seconds of CPU time on a PMac dual 550-MHz G4
	private double IntenForward[] = new double[256];
	private static final int InvGamTblSize = 65536;
	private int IntenInverse[] = new int[InvGamTblSize];
	
	// Input and output values
	private double[] ColorIn = new double[3];
	private double[] ColorOut = new double[3];
		
	// Get the parameter values:
	public int getType() {return cbtype;}
	public double getGamma() {return gamma;}
	
	public void setType(int cbtype_)
	{
		cbtype = cbtype_;
		SetTransformType(cbtype);
	}
		
	public void setGamma(double gamma_)
	{
		gamma = gamma_;
		
		if (gamma == 1)
		{
			// Fast calculation for this case
			for (int i=0; i<256; i++)
			{
				double val = (i+0.5)/256.0;
				IntenForward[i] = val;
			}
			
			for (int i=0; i<InvGamTblSize; i++)
			{
				int ival = (256*i)/InvGamTblSize;
				if (ival > 0xff) ival = 0xff;
				IntenInverse[i] = ival;
			}
		}
		else
		{
			for (int i=0; i<256; i++)
			{
				double val = (i+0.5)/256.0;
				val =  Math.pow(val, gamma);
				IntenForward[i] = val;
			}
			
			double invgamma = 1/gamma;
			for (int i=0; i<InvGamTblSize; i++)
			{
				double val = (i+0.5)/((double)InvGamTblSize);
				val = Math.pow(val, invgamma);
				int ival = (int)(256.0*val);
				if (ival > 0xff) ival = 0xff;
				IntenInverse[i] = ival;
			}
		}
	}
	
	// Keep the constructor simple
	CB_Filter()
	{
		setGamma(1);
	}
	
	// Filtering function
	public int filterRGB(int x, int y, int rgb)
	{
		// Peel off the color values
		ColorIn[0] = IntenForward[(rgb >> 16) & 0xff];
		ColorIn[1] = IntenForward[(rgb >> 8) & 0xff];
		ColorIn[2] = IntenForward[(rgb) & 0xff];
		
		DoTransform(ColorIn,ColorOut);
		
		// Take care of the fractional case
		if (fraction < 1)
			for (int i=0; i<3; i++)
				ColorOut[i] = ColorIn[i] + fraction*(ColorOut[i] - ColorIn[i]);
		
		// Handle off-edge:
		if (off_edge > 0)
		{
			// Push to grayscale while keeping the hue
			double gray = 0.299*ColorIn[0] + 0.587*ColorIn[1] + 0.114*ColorIn[2];
			if (gray > 0)
			{
				// Offset amount >= 1: don't push
				double offset = 1;
				for (int i=0; i<3; i++)
				{
					if (ColorOut[i] < 0)
					{
						double iofst = gray/(gray - ColorOut[i]);
						if (iofst < offset) offset = iofst;
					}
					else if (ColorOut[i] >= 1)
					{
						double iofst = (1 - gray)/(ColorOut[i] - gray);
						if (iofst < offset) offset = iofst;
					}
				}
				if (offset < 1)
					for (int i=0; i<3; i++)
						ColorOut[i] = gray + offset*(ColorOut[i] - gray);
			}
		}
		
		// Stuff them in; clamp to range if necessary
		int res = 0xff;
		for (int i=0; i<3; i++)
		{
			int ival = (int)(InvGamTblSize*ColorOut[i]);
			if (ival < 0) ival = 0;
			else if (ival >= InvGamTblSize) ival = (InvGamTblSize-1);
			res = (res << 8) + IntenInverse[ival];
		}
		return res;
	}
	
	// Work on an image; return a filtered version
	public BufferedImage filter(Image img)
	{
		// Use it -- to use the AWT Toolkit, create a Canvas lightweight dummy object,
		// which can be created in headless mode
		Canvas dummy = new Canvas();
		Image fltimg = dummy.createImage(new FilteredImageSource(img.getSource(), this));
		
		// Create a buffered image from a generic image object
		BufferedImage outimg =
			new BufferedImage(img.getWidth(null), img.getHeight(null), BufferedImage.TYPE_INT_RGB);
		Graphics2D g = outimg.createGraphics();
		g.drawImage(fltimg, 0, 0, null);
		
		return outimg;
	}
}
