//
//  ImageWindow.java
//  CBSim
//
//  Created by Loren Petrich on 5/18/08.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.imageio.*;
import javax.swing.*;

class ImageWindow extends JFrame implements ListenerSetType
{
	// Where the image is displayed
	private final ImagePanel ImgPnl;
	
	// Image title, to keep it from being duplicated
	// Base version -- use for making other instances
	private String TitleBase;
	// With image size (will have color-blindness type added)
	private String TitleBaseExtra;
	
	// Share instances of these
	private static ImageOpenFileHandler IOFH = new ImageOpenFileHandler();
	private static ImageSaveFileHandler ISFH  = new ImageSaveFileHandler();
	private static TimerDialog TmrDlg = new TimerDialog();
	
	private Vector<JRadioButtonMenuItem> TypeSels;
	
	// Static factory methods; they return new null if they fail
	static ImageWindow Make()
	{
		URL ResURL;
		try
		{
			// Kludgy way of getting a static method's class object,
			// since we can't do this.getClass() in it.
			Exception ex = new Exception();
			StackTraceElement[] STEList = ex.getStackTrace();
			String ClassName = STEList[0].getClassName();
			Class Cls = Class.forName(ClassName);
			
			ResURL = Cls.getResource("CBSim.png");
		}
		catch(Exception ex)
		{
			ShowError("Class-name error: " + ex.toString());
			return null;
		}
		return Make(ResURL, "(Demo Image)", null);
	}
	
	static ImageWindow Make(File file)
	{
		return Make(file, file.getName(), file.toString());
	}
	
	static ImageWindow Make(File file, String Title, String FullTitle)
	{
		BufferedImage Img;
		try
		{
			if (!file.exists())
			{
				ShowError("File does not exist: " + file);
				return null;
			}
			if (!file.isFile())
			{
				ShowError("Is not a file: " + file);
				return null;
			}
			if (!file.canRead())
			{
				ShowError("File is not readable: " + file);
				return null;
			}
			
			Img = ImageIO.read(file);
		}
		catch (Exception ex)
		{
			ShowError("Image-from-file error: " + ex.toString());
			return null;
		}
		
		return MakeFromImage(Img, Title, FullTitle, "Image not read from file: " + file.toString());
	}
	
	static ImageWindow Make(URL url)
	{
		// Make a title:
		String urlstr = url.toString();
		int LastSepLoc = urlstr.lastIndexOf('/');
		String urlttl = urlstr.substring(LastSepLoc+1);
		
		return Make(url, urlttl, urlstr);
	}
	
	static ImageWindow Make(URL url, String Title, String FullTitle)
	{
		BufferedImage Img;
		try
		{
			Img = ImageIO.read(url);
		}
		catch(Exception ex)
		{
			ShowError("Image-from-URL error: " + ex.toString());
			return null;
		}
		
		return MakeFromImage(Img, Title, FullTitle, "Image not read from URL: " + url.toString());
	}
	
	static ImageWindow MakeFromFile(String filestr)
	{
		return Make(new File(filestr));
	}
	
	static ImageWindow MakeFromFile(String filestr, String Title, String FullTitle)
	{
		return Make(new File(filestr), Title, FullTitle);
	}
	
	static ImageWindow MakeFromURL(String urlstr)
	{
		URL url = MakeURL(urlstr);
		if (url == null) return null;
		return Make(url);
	}
	
	static ImageWindow MakeFromURL(String urlstr, String Title, String FullTitle)
	{
		URL url = MakeURL(urlstr);
		if (url == null) return null;
		return Make(url, Title, FullTitle);
	}
	
	static ImageWindow MakeFromClipboard()
	{
		Image Img = ImageClipboardHandler.GetImage();
		if (Img == null) return null;
		
		BufferedImage BImg = MakeBuffered(Img);
		return MakeFromImage(BImg, "(Clipboard Image)", null, "Image not read from clipboard");
	}
	
	static ImageWindow MakeFromScreenshot()
	{
		// If unconfirmed, then bug out
		if (!TmrDlg.show()) return null;
		
		BufferedImage BImg = ImageFromScreenshot.Get();

		return MakeFromImage(BImg, "(Screenshot Image)", null, "Image not read from screenshot");
	}
	
	// Constructor
	ImageWindow(BufferedImage Img, String Title, String FullTitle)
	{
		TitleBase = Title;
		TitleBaseExtra = String.format("%s - %d x %d", TitleBase, Img.getWidth(), Img.getHeight());
		
		ImgPnl = new ImagePanel(Img, this, FullTitle);
		add(ImgPnl);
		
		JMenuItem Item;
		
		JMenu FileMenu = new JMenu("File");
		
		MakeMenuItem(FileMenu, "New from Clipboard", 'N',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ImageWindow.MakeFromClipboard();
				}
			}
		);

		MakeMenuItem(FileMenu, "Open File...", 'O',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					File file = IOFH.GetFile();
					if (file != null)
						Make(file);
				}
			}
		);
		
		MakeMenuItem(FileMenu, "Open URL...", 'L',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					String urlstr = JOptionPane.showInputDialog(null,"What image URL?");
					if (urlstr != null)
					{
						MakeFromURL(urlstr);
					}
				}
			}
		);

		MakeMenuItem(FileMenu, "Show Demo Image",
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ImageWindow.Make();
				}
			}
		);

		MakeMenuItem(FileMenu, "Take Screenshot...",
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ImageWindow.MakeFromScreenshot();
				}
			}
		);
		
		FileMenu.addSeparator();
		
		MakeMenuItem(FileMenu, "Duplicate", 'D',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					new ImageWindow(ImgPnl.OrigImg, TitleBase, ImgPnl.GetTitle());
				}
			}
		);
		
		FileMenu.addSeparator();
		
		MakeMenuItem(FileMenu, "Save As...", 'S',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					try
					{
						ISFH.SaveImage(ImgPnl.ResImg);
					}
					catch(Exception ex)
					{
						ShowError("Error in writing out image: " + ex.toString());
					}
				}
			}
		);
		
		FileMenu.addSeparator();
		
		MakeMenuItem(FileMenu, "Close", 'W',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					dispose();
				}
			}
		);
		
		JMenu EditMenu = new JMenu("Edit");
		
		Item = MakeMenuItem(EditMenu, "Undo", 'Z', null);
		Item.setEnabled(false);
		
		Item = MakeMenuItem(EditMenu, "Cut", 'X', null);
		Item.setEnabled(false);
		
		MakeMenuItem(EditMenu, "Copy", 'C',
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ImageClipboardHandler.PutImage(ImgPnl.ResImg);
				}
			}
		);
		
		Item = MakeMenuItem(EditMenu, "Copy Filename/URL",
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					String PanelTitle = ImgPnl.GetTitle();
					if (PanelTitle.length() > 0)
						ImageClipboardHandler.PutText(PanelTitle);
				}
			}
		);
		
		// We don't need this if the filename/URL is not present
		if (FullTitle == null)
			Item.setEnabled(false);
		else if (FullTitle.length() == 0)
			Item.setEnabled(false);
		
		Item = MakeMenuItem(EditMenu, "Paste", 'V', null);
		Item.setEnabled(false);
		
		JMenu ViewMenu = new JMenu("View");
		
		ActionListener SBAL = new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				String ac = e.getActionCommand();
				int it = 0;
				try
				{
					it = Integer.valueOf(ac);
				}
				catch (Exception ex)
				{
					return;
				}
				ImgPnl.SetType(it);
			}
		};
		
		TypeSels = new Vector<JRadioButtonMenuItem>();
		ButtonGroup TypeGroup = new ButtonGroup();
		for (int it=0; it<ImagePanel.NumTypes; it++)
		{
			JRadioButtonMenuItem RBItem;
			RBItem = MakeRadioButtonMenuItem(ViewMenu, ImagePanel.OptionNames[it], '0'+it, SBAL);
			RBItem.setActionCommand(Integer.toString(it));
			TypeGroup.add(RBItem);
			TypeSels.add(RBItem);
		}
		
		SetType(ImgPnl.GetType());

		ViewMenu.addSeparator();		
		
		MakeMenuItem(ViewMenu, "Extras...",
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ImgPnl.ShowExtrasDialog();
				}
			}
		);
		
		JMenuBar MenuBar = new JMenuBar();
		MenuBar.add(FileMenu);
		MenuBar.add(EditMenu);
		MenuBar.add(ViewMenu);
		setJMenuBar(MenuBar);
		
		pack();
		setVisible(true);
	}
	
	public void SetType(int whch)
	{
		JRadioButtonMenuItem RBItem = TypeSels.get(whch);
		RBItem.setSelected(true);
		setTitle(String.format("%s - %s", TitleBaseExtra, ImagePanel.OptionNames[whch]));
	}
	
	static URL MakeURL(String urlstr)
	{
		URL url;
		try
		{
			url = new URL(urlstr);
		}
		catch(Exception ex)
		{
			ShowError("Bad image URL: " + ex.toString());
			return null;
		}
		return url;
	}
	
	static JMenuItem MakeMenuItem(JMenu theMenu, String Text, ActionListener Listener)
	{
		JMenuItem Item = new JMenuItem(Text);
		theMenu.add(Item);
		Item.addActionListener(Listener);
		return Item;
	}
	
	static JMenuItem MakeMenuItem(JMenu theMenu, String Text, int Shortcut, ActionListener Listener)
	{
		JMenuItem Item = MakeMenuItem(theMenu, Text, Listener);
		Item.setAccelerator(KeyStroke.getKeyStroke(Shortcut,
			Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
		return Item;
	}
	
	static JRadioButtonMenuItem MakeRadioButtonMenuItem(JMenu theMenu, String Text, ActionListener Listener)
	{
		JRadioButtonMenuItem Item = new JRadioButtonMenuItem(Text);
		theMenu.add(Item);
		Item.addActionListener(Listener);
		return Item;
	}
	
	static JRadioButtonMenuItem MakeRadioButtonMenuItem(JMenu theMenu, String Text, int Shortcut, ActionListener Listener)
	{
		JRadioButtonMenuItem Item = MakeRadioButtonMenuItem(theMenu, Text, Listener);
		Item.setAccelerator(KeyStroke.getKeyStroke(Shortcut,
			Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
		return Item;
	}
	
	static ImageWindow MakeFromImage(BufferedImage Img, String Title, String FullTitle, String Error)
	{
		if (Img == null)
		{
			ShowError(Error);
			return null;
		}
		return new ImageWindow(Img, Title, FullTitle);
	}
	
	static BufferedImage MakeBuffered(Image Img)
	{
		BufferedImage BImg = new BufferedImage(Img.getWidth(null), Img.getHeight(null),
			BufferedImage.TYPE_INT_RGB);
		Graphics2D g = BImg.createGraphics();
		g.drawImage(Img, 0, 0, null);
		return BImg;
	}
	
	static void ShowError(String ErrStr)
	{
		JOptionPane.showMessageDialog(null, ErrStr, "CBSim Message", JOptionPane.ERROR_MESSAGE);
	}
}